"""Exceptions."""
from __future__ import annotations

from typing import Literal, cast

import aiohttp


class APIError(Exception):
    message: str
    request: aiohttp.ClientRequest

    body: object | None
    """The API response body.

    If the API responded with a valid JSON structure then this property will be the
    decoded result.

    If it isn't a valid JSON structure then this will be the raw response.

    If there was no response associated with this error then it will be `None`.
    """

    code: str | None
    param: str | None
    type: str | None

    def __init__(self, message: str, request: aiohttp.ClientRequest | aiohttp.RequestInfo, *,
                 body: object | None) -> None:
        super().__init__(message)
        self.request = request
        self.message = message

        if isinstance(body, dict):
            self.code = cast(any, body.get("code"))
            self.param = cast(any, body.get("param"))
            self.type = cast(any, body.get("type"))
        else:
            self.code = None
            self.param = None
            self.type = None


class APIStatusError(APIError):
    """Raised when an API response has a status code of 4xx or 5xx."""

    response: aiohttp.ClientResponse
    status_code: int

    def __init__(self, message: str, *, response: aiohttp.ClientResponse, body: object | None) -> None:
        super().__init__(message, response.request_info, body=body)
        self.response = response
        self.status_code = response.status


class APIConnectionError(APIError):
    def __init__(self, *, message: str = "Connection error.",
                 request: aiohttp.ClientRequest | aiohttp.RequestInfo) -> None:
        super().__init__(message, request, body=None)


class APITimeoutError(APIConnectionError):
    def __init__(self, request: aiohttp.ClientRequest | aiohttp.RequestInfo) -> None:
        super().__init__(message="Request timed out.", request=request)


class BadRequestError(APIStatusError):
    status_code: Literal[400] = 400


class AuthenticationError(APIStatusError):
    status_code: Literal[401] = 401


class PermissionDeniedError(APIStatusError):
    status_code: Literal[403] = 403


class NotFoundError(APIStatusError):
    status_code: Literal[404] = 404


class ConflictError(APIStatusError):
    status_code: Literal[409] = 409


class UnprocessableEntityError(APIStatusError):
    status_code: Literal[422] = 422


class RateLimitError(APIStatusError):
    status_code: Literal[429] = 429


class InternalServerError(APIStatusError):
    pass


class HttpException(Exception):  # noqa: N818
    """Exception base for HTTP errors."""

    def __init__(
        self,
        status: int,
        message: str = "HTTP error",
        errors: dict[str, str] | None = None,
    ):
        self.status = status
        self.message = message
        self.errors = errors
        super().__init__(self.message)


class FatalHttpException(HttpException):
    """Exception raised for HTTP codes that are non-retryable."""


class RetryableHttpException(HttpException):
    """Exception raised for HTTP codes that are possible to retry."""


class AuthorizationError(FatalHttpException):
    """Invalid login exception."""



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from typing import Literal

from pydantic import BaseModel
from typing_extensions import TypedDict

from .typing import StrEnum

Unit = Literal[
    "cl", "cm", "dl", "l",
    "g", "hg", "kg",
    "m", "m100", "ml",
    "pair", "dosage",
    "piece", "portion", "squareMeter"
]


class PhysicalStoreGroup(StrEnum):
    MENY_NO = "MENY_NO"
    SPAR_NO = "SPAR_NO"
    JOKER_NO = "JOKER_NO"
    ODA_NO = "ODA_NO"
    ENGROSSNETT_NO = "ENGROSSNETT_NO"
    NAERBUTIKKEN = "NAERBUTIKKEN"
    BUNNPRIS = "BUNNPRIS"
    KIWI = "KIWI"
    REMA_1000 = "REMA_1000"
    EUROPRIS_NO = "EUROPRIS_NO"
    HAVARISTEN = "HAVARISTEN"
    HOLDBART = "HOLDBART"
    FUDI = "FUDI"
    COOP_NO = "COOP_NO"
    COOP_MARKED = "COOP_MARKED"
    MATKROKEN = "MATKROKEN"
    COOP_MEGA = "COOP_MEGA"
    COOP_PRIX = "COOP_PRIX"
    COOP_OBS = "COOP_OBS"
    COOP_EXTRA = "COOP_EXTRA"
    COOP_BYGGMIX = "COOP_BYGGMIX"
    COOP_OBS_BYGG = "COOP_OBS_BYGG"
    COOP_ELEKTRO = "COOP_ELEKTRO"
    ARK = "ARK"
    NORLI = "NORLI"
    ADLIBRIS = "ADLIBRIS"


class KassalappResource(BaseModel, ABC):
    """Kassalapp resource."""

    id: int | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None


@dataclass
class AllergenItem:
    code: str
    display_name: str
    contains: str


class Icon(TypedDict):
    svg: str | None
    png: str | None


@dataclass
class LabelItem:
    name: str | None
    display_name: str | None
    description: str | None
    organization: str | None
    alternative_names: str | None
    type: str | None
    year_established: int | None
    about: str | None
    note: str | None
    icon: Icon | None


@dataclass
class NutritionItem:
    code: str
    display_name: str
    amount: float
    unit: str


class OpeningHours(TypedDict):
    monday: str | None
    tuesday: str | None
    wednesday: str | None
    thursday: str | None
    friday: str | None
    saturday: str | None
    sunday: str | None


class Position(TypedDict):
    lat: float
    lng: float


class ProximitySearch(Position):
    km: float


class PhysicalStore(KassalappResource):
    id: int | None
    group: PhysicalStoreGroup | None
    name: str | None
    address: str | None
    phone: str | None
    email: str | None
    fax: str | None
    logo: str | None
    website: str | None
    detailUrl: str | None
    position: Position | None
    openingHours: OpeningHours | None


class ProductCategory(TypedDict):
    id: int
    depth: int
    name: str


class Store(TypedDict):
    name: str
    code: str | None
    url: str | None
    logo: str | None


class Price(TypedDict):
    price: float
    date: datetime


class CurrentPrice(Price):
    unit_price: float | None


class ProductBase(KassalappResource):
    name: str | None = None
    vendor: str | None = None
    brand: str | None = None
    description: str | None = None
    ingredients: str | None = None
    url: str | None = None
    image: str | None = None
    category: list[ProductCategory] | None = None
    store: Store | None = None
    weight: float | None = None
    weight_unit: Unit | None = None
    price_history: list[Price] | None = None


class Product(ProductBase):
    ean: int | None
    current_price: float | None
    current_unit_price: float | None
    allergens: list[AllergenItem] | None
    nutrition: list[NutritionItem] | None
    labels: list[LabelItem] | None


class ProductComparisonItem(ProductBase):
    current_price: CurrentPrice | None
    kassalapp: dict[str, str]


class ProductComparison(ProductBase):
    ean: int | None
    products: list[ProductComparisonItem] | None
    allergens: list[AllergenItem] | None
    nutrition: list[NutritionItem] | None
    labels: list[LabelItem] | None


class ShoppingListItem(KassalappResource):
    text: str | None
    checked: bool
    product: ProductComparison | None


class ShoppingList(KassalappResource):
    title: str
    items: list[ShoppingListItem] | None = None


class MessageResponse(BaseModel):
    message: str


class StatusResponse(BaseModel):
    status: str


class Webhook(KassalappResource):
    name: str | None = None
    url: str
    secret: str | None = None
    eans: list[str] | None = None
    ids: list[str] | None = None



from __future__ import annotations

import logging
from enum import Enum
_LOGGER = logging.getLogger(__name__)


# noinspection PyUnresolvedReferences
class StrEnum(str, Enum):
    """A string enumeration of type `(str, Enum)`.
    All members are compared via `upper()`. Defaults to UNKNOWN.
    """
    def __eq__(self, other: str) -> bool:
        other = other.upper()
        return super().__eq__(other)

    @classmethod
    def _missing_(cls, value) -> str:
        has_unknown = False
        for member in cls:
            if member.name.upper() == "UNKNOWN":
                has_unknown = True
            if member.name.upper() == value.upper():
                return member
        if has_unknown:
            _LOGGER.warning("'%s' is not a valid '%s'", value, cls.__name__)
            return cls.UNKNOWN
        raise ValueError(f"'{value}' is not a valid {cls.__name__}")



"""Constants used by kassalapp."""
from http import HTTPStatus
from importlib.metadata import version
from typing import Final

VERSION = version("kassalappy")

API_ENDPOINT: Final = "https://kassal.app/api/v1"
DEFAULT_TIMEOUT: Final = 10

API_ERR_CODE_UNKNOWN: Final = "UNKNOWN"
HTTP_CODES_RETRYABLE: Final = [
    HTTPStatus.TOO_MANY_REQUESTS,
    HTTPStatus.PRECONDITION_REQUIRED,
]
HTTP_CODES_NO_ACCESS: Final = [
    HTTPStatus.UNAUTHORIZED,
    HTTPStatus.FORBIDDEN,
]
HTTP_CODES_FATAL: Final = [HTTPStatus.BAD_REQUEST]



"""Kassalapp CLI."""
from __future__ import annotations

import logging

import asyncclick as click
from tabulate import tabulate

from kassalappy import Kassalapp
from kassalappy.models import PhysicalStoreGroup, ProximitySearch

TABULATE_DEFAULTS = {
    "tablefmt": "rounded_grid",
}


@click.group()
@click.password_option("--token", type=str, required=True, confirmation_prompt=False, help="API Token")
@click.option("--debug", is_flag=True, help="Set logging level to DEBUG")
@click.pass_context
async def cli(ctx: click.Context, token: str, debug: bool):
    """Kassalapp CLI."""
    configure_logging(debug)
    client = Kassalapp(access_token=token)

    ctx.ensure_object(dict)
    ctx.obj["DEBUG"] = debug
    ctx.obj["client"] = client
    await ctx.with_async_resource(client)


@cli.command("health")
@click.pass_context
async def health(ctx: click.Context):
    """Checks if the Kassalapp API is working."""
    client: Kassalapp = ctx.obj["client"]
    data = await client.healthy()
    click.echo(data)


@cli.command("shopping-lists")
@click.option("--items", is_flag=True, help="Include shopping list items")
@click.pass_context
async def shopping_lists(ctx: click.Context, items: bool):
    """Get shopping lists associated with the authenticated user."""
    client: Kassalapp = ctx.obj["client"]
    data = await client.get_shopping_lists(include_items=items)
    click.echo(tabulate([m.model_dump() for m in data], headers="keys", **TABULATE_DEFAULTS))


@cli.command("shopping-list")
@click.argument("list_id", type=int)
@click.option("--items", is_flag=True, help="Include shopping list items")
@click.pass_context
async def shopping_list(ctx: click.Context, list_id: int, items: bool):
    """Get details for a specific shopping list."""
    client: Kassalapp = ctx.obj["client"]
    data = await client.get_shopping_list(list_id, include_items=items)
    data_model = data.model_dump()
    exclude_keys = ['items']
    new_d = {k: data_model[k] for k in set(list(data_model.keys())) - set(exclude_keys)}
    click.echo(tabulate([new_d], headers="keys", **TABULATE_DEFAULTS))
    click.echo("Items")
    click.echo(tabulate(data_model['items'], headers="keys", **TABULATE_DEFAULTS))


@cli.command("add-item")
@click.option("--list_id", type=int)
@click.argument("text", required=True)
@click.argument("product_id", type=int, required=False, default=None)
@click.pass_context
async def add_item(ctx: click.Context, list_id: int, text: str, product_id: int | None = None):
    """Add an item to shopping list."""
    client: Kassalapp = ctx.obj["client"]
    response = await client.add_shopping_list_item(list_id, text, product_id)
    click.echo(response)


@cli.command("product")
@click.argument("search", type=str)
@click.option("--count", type=int, default=5, help="Number of results to return")
@click.pass_context
async def product_search(ctx: click.Context, search: str, count: int):
    """Search for products."""
    client: Kassalapp = ctx.obj["client"]
    results = await client.product_search(search=search, size=count, unique=True)
    click.echo(tabulate([r.model_dump() for r in results], headers="keys", **TABULATE_DEFAULTS))


@cli.command("store-groups")
async def store_groups():
    """Get list of available physical store groups."""
    groups = [g.value for g in PhysicalStoreGroup]
    groups.sort()
    click.echo(groups)


@cli.command("stores")
@click.option("--proximity",
              type=ProximitySearch,
              help="Proximity of stores to search for (latitude longitude radius)")
@click.option("--group", type=PhysicalStoreGroup)
@click.argument("search", type=str, required=False)
@click.option("--count", type=int, help="Number of results to return")
@click.pass_context
async def store_search(
    ctx: click.Context,
    proximity: ProximitySearch,
    group: PhysicalStoreGroup,
    search: str,
    count: int,
):
    """Search for physical stores."""
    proximity_search = None
    if proximity:
        lat, lng, radius = proximity
        proximity_search = ProximitySearch(lat=lat, lng=lng, km=radius)

    client: Kassalapp = ctx.obj["client"]
    results = await client.physical_stores(search=search, group=group, proximity=proximity_search, size=count)
    click.echo(tabulate([r.model_dump() for r in results], headers="keys", **TABULATE_DEFAULTS))


@cli.command("webhooks")
@click.pass_context
async def webhooks(ctx: click.Context):
    """Get webhooks."""
    client: Kassalapp = ctx.obj["client"]
    results = await client.get_webhooks()
    click.echo(tabulate([r.model_dump() for r in results], headers="keys", **TABULATE_DEFAULTS))


def configure_logging(debug: bool):
    """Set up logging."""
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level)



"""Library to handle connection with kassalapp web API."""
from __future__ import annotations

import json
import logging
from http import HTTPStatus
from types import NoneType
from typing import Literal, TypeVar, cast

import aiohttp
from pydantic import ValidationError

from .const import (
    API_ENDPOINT,
    DEFAULT_TIMEOUT,
    VERSION,
)
from .exceptions import (
    APIStatusError,
    APITimeoutError,
    AuthenticationError,
    BadRequestError,
    ConflictError,
    InternalServerError,
    NotFoundError,
    PermissionDeniedError,
    RateLimitError,
    UnprocessableEntityError,
)
from .models import (
    KassalappResource,
    MessageResponse,
    PhysicalStore,
    PhysicalStoreGroup,
    Product,
    ProductComparison,
    ProximitySearch,
    ShoppingList,
    ShoppingListItem,
    StatusResponse,
    Webhook,
)

R = TypeVar("R")

ResponseT = TypeVar(
    "ResponseT",
    bound=KassalappResource | MessageResponse | StatusResponse | dict[str, any],
)

_LOGGER = logging.getLogger(__name__)


# noinspection PyMethodMayBeStatic
class Kassalapp:
    """Class to communicate with the Kassalapp API."""

    def __init__(
        self,
        access_token: str,
        timeout: int = DEFAULT_TIMEOUT,
        websession: aiohttp.ClientSession | None = None,
    ):
        """Initialize the Kassalapp connection."""
        self._user_agent: str = f"python kassalappy/{VERSION}"
        self.websession = websession
        self.timeout: int = timeout
        self._access_token: str = access_token

    async def __aenter__(self) -> Kassalapp:
        if self.websession is None:
            self.websession = aiohttp.ClientSession()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.websession.close()

    async def _process_response(
        self,
        cast_to: type[ResponseT],
        response: aiohttp.ClientResponse,
    ) -> R:
        response_data = await response.json()

        if response.ok:
            data = response_data.get("data") or response_data
        else:
            data = response_data

        try:
            return await self._process_response_data(
                data=data,
                cast_to=cast_to,
            )
        except ValidationError:
            _LOGGER.exception("Error validating response data")
            raise

    async def _process_response_data(
        self,
        data: object | list[object],
        cast_to: type[ResponseT],
    ) -> ResponseT | list[ResponseT]:
        if cast_to is NoneType:
            return cast(R, None)

        if cast_to == str:
            return cast(R, data)

        if data is None:
            return cast(ResponseT, None)

        if type(data) is list:
            return [cast(ResponseT, cast_to.model_validate(d)) for d in data]

        return cast(ResponseT, cast_to.model_validate(data))

    def _make_status_error_from_response(
        self,
        response: aiohttp.ClientResponse,
        err_text: str
    ) -> APIStatusError:
        body = err_text.strip()
        try:
            body = json.loads(err_text)
            err_msg = f"Error code: {response.status} - {body}"
        except Exception:  # noqa: BLE001
            err_msg = err_text or f"Error code: {response.status}"

        return self._make_status_error(err_msg, body=body, response=response)

    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: aiohttp.ClientResponse,
    ) -> APIStatusError:
        if response.status == HTTPStatus.BAD_REQUEST:
            return BadRequestError(err_msg, response=response, body=body)

        if response.status == HTTPStatus.UNAUTHORIZED:
            return AuthenticationError(err_msg, response=response, body=body)

        if response.status == HTTPStatus.FORBIDDEN:
            return PermissionDeniedError(err_msg, response=response, body=body)

        if response.status == HTTPStatus.NOT_FOUND:
            return NotFoundError(err_msg, response=response, body=body)

        if response.status == HTTPStatus.CONFLICT:
            return ConflictError(err_msg, response=response, body=body)

        if response.status == HTTPStatus.UNPROCESSABLE_ENTITY:
            return UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status == HTTPStatus.TOO_MANY_REQUESTS:
            return RateLimitError(err_msg, response=response, body=body)

        if response.status >= HTTPStatus.INTERNAL_SERVER_ERROR:
            return InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)

    async def execute(
        self,
        endpoint: str,
        cast_to: type[ResponseT],
        method: str = aiohttp.hdrs.METH_GET,
        params: dict[str, any] | None = None,
        data: dict[any, any] | None = None,
        timeout: int | None = None,
    ) -> ResponseT:
        """"Execute a API request and return the data."""
        timeout = timeout or self.timeout

        request_args = {
            "headers": {
                "Authorization": "Bearer " + self._access_token,
                aiohttp.hdrs.USER_AGENT: self._user_agent,
            },
            "json": data,
            "params": params,
            "timeout": aiohttp.ClientTimeout(total=timeout),
        }
        request_url = f"{API_ENDPOINT}/{endpoint}"
        try:
            async with self.websession as session:
                response = await session.request(method, request_url, **request_args)
                body = await response.text()
                response.raise_for_status()
        except aiohttp.ServerTimeoutError as err:
            raise APITimeoutError(request=response.request_info) from err
        except (aiohttp.ClientResponseError, aiohttp.ClientError):
            raise self._make_status_error_from_response(response, body) from None

        return await self._process_response(
            cast_to=cast_to,
            response=response,
        )

    async def healthy(self):
        """Check if the Kassalapp API is working."""
        return await self.execute("health", StatusResponse)

    async def get_shopping_lists(self, include_items: bool = False):
        """Get shopping lists."""
        params = {}
        if include_items:
            params["include"] = "items"
        return await self.execute("shopping-lists", ShoppingList, params=params)

    async def get_shopping_list(self, list_id: int, include_items: bool = False):
        """Get a shopping list."""
        params = {}
        if include_items:
            params["include"] = "items"
        return await self.execute(f"shopping-lists/{list_id}", ShoppingList, params=params)

    async def create_shopping_list(self, title: str):
        """Create a new shopping list."""
        return await self.execute("shopping-lists", ShoppingList, "post", data={"title": title})

    async def delete_shopping_list(self, list_id: int):
        """Delete a shopping list."""
        return await self.execute(f"shopping-lists/{list_id}", ShoppingList, "delete")

    async def update_shopping_list(self, list_id: int, title: str):
        """Update a new shopping list."""
        return await self.execute(f"shopping-lists/{list_id}", ShoppingList, "patch", data={"title": title})

    async def get_shopping_list_items(self, list_id: int):
        """Shorthand method to get all items from a shopping list."""
        shopping_list = await self.get_shopping_list(list_id, include_items=True)
        return shopping_list.get("items", [])

    async def add_shopping_list_item(self, list_id: int, text: str, product_id: int | None = None):
        """Add an item to an existing shopping list."""
        item = {
            "text": text,
            "product_id": product_id,
        }
        return await self.execute(f"shopping-lists/{list_id}/items", ShoppingListItem, "post", data=item)

    async def delete_shopping_list_item(self, list_id: int, item_id: int):
        """Remove an item from the shopping list."""
        return await self.execute(
            f"shopping-lists/{list_id}/items/{item_id}",
            MessageResponse,
            "delete",
        )

    async def update_shopping_list_item(
        self,
        list_id: int,
        item_id: int,
        text: str | None = None,
        checked: bool | None = None,
    ):
        """Update an item in the shopping list."""
        data = {
            "text": text,
            "checked": checked,
        }
        return await self.execute(
            f"shopping-lists/{list_id}/items/{item_id}",
            ShoppingListItem,
            "patch",
            data={k: v for k, v in data.items() if v is not None},
        )

    async def product_search(
        self,
        search: str | None = None,
        brand: str | None = None,
        vendor: str | None = None,
        excl_allergens: list[str] | None = None,
        incl_allergens: list[str] | None = None,
        exclude_without_ean: bool = False,
        price_max: float | None = None,
        price_min: float | None = None,
        size: int | None = None,
        sort: Literal["date_asc", "date_desc", "name_asc", "name_desc", "price_asc", "price_desc"] | None = None,
        unique: bool = False,
    ):
        """Search for groceries and various product to find price, ingredients and nutritional information.

        :param search: Search for products based on a keyword.
                       The keyword must be a string with a minimum length of 3 characters.
        :param brand: Filter products by brand name.
        :param vendor: Filter products by vendor (leverandør).
        :param excl_allergens: Exclude specific allergens from the products.
        :param incl_allergens: Include only specific allergens in the products.
        :param exclude_without_ean: If true, products without an EAN number are excluded from the results.
        :param price_max: Filter products by maximum price.
        :param price_min: Filter products by minimum price.
        :param size: The number of products to be displayed per page.
                     Must be an integer between 1 and 100.
        :param sort: Sort the products by a specific criteria.
        :param unique: If true, the product list will be collapsed based on the EAN number of the product;
                       in practice, set this to true if you don't want duplicate results.
        :return:
        """
        params = {
            "search": search,
            "brand": brand,
            "vendor": vendor,
            "excl_allergens": excl_allergens,
            "incl_allergens": incl_allergens,
            "exclude_without_ean": 1 if exclude_without_ean is True else None,
            "price_min": price_min,
            "price_max": price_max,
            "size": size,
            "sort": sort,
            "unique": 1 if unique is True else None,
        }

        return await self.execute(
            "products",
            Product,
            params={k: v for k, v in params.items() if v is not None},
        )

    async def product_find_by_url(self, url: str):
        """Will look up product information based on a URL.
        Returns the product price, nutritional information, ingredients, allergens for the product.
        """
        params = {
            "url": url,
        }
        return await self.execute(
            "products/find-by-url/single",
            Product,
            params=params,
        )

    async def products_find_by_url(self, url: str):
        """Will look up product information based on a URL.
        Returns all matching prices from other stores that stock that item.
        """
        params = {
            "url": url,
        }
        return await self.execute(
            "products/find-by-url/compare",
            ProductComparison,
            params=params,
        )

    async def product_get_by_id(self, product_id: int):
        """Gets a specific product by id."""
        return await self.execute(
            f"products/id/{product_id}",
            Product,
        )

    async def product_get_by_ean(self, ean: str):
        """Gets a specific product by EAN (barcode) number."""
        return await self.execute(
            f"products/ean/{ean}",
            ProductComparison,
        )

    async def physical_stores(
        self,
        search: str | None = None,
        group: PhysicalStoreGroup | None = None,
        proximity: ProximitySearch | None = None,
        size: int | None = None,
    ):
        """Search for physical stores.

        Useful for finding a grocery store by name, location or based on the group (grocery store chain),
        returns name, address, contact information and opening hours for each store.

        :param search: Perform a search based on a keyword.
        :param group: Filter by group name.
        :param proximity: Search radius in kilometers for proximity search.
        :param size: The number of results to be displayed per page. Must be an integer between 1 and 100.
        :return:
        """
        params = {
            "search": search,
            "group": group.value if group is not None else None,
            "size": size,
        }
        if proximity is not None:
            params['km'] = proximity['km']
            params['lat'] = proximity['lat']
            params['lng'] = proximity['lng']

        return await self.execute(
            "physical-stores",
            PhysicalStore,
            params={k: v for k, v in params.items() if v is not None},
        )

    async def physical_store(self, store_id: int):
        """Finds a grocery store by ID."""
        return await self.execute(
            f"physical-stores/{store_id}",
            PhysicalStore,
        )

    async def get_webhooks(self):
        """Retrieves a collection of webhooks associated with the authenticated user."""
        return await self.execute("webhooks", Webhook)

    async def create_webhook(
        self,
        url: str,
        name: str | None = None,
        ids: list[str] | None = None,
        eans: list[str] | None = None,
    ):
        """Create and store a new webhook associated with the authenticated user."""
        params = {
            "url": url,
            "name": name,
            "ids": ids,
            "eans": eans,
        }
        return await self.execute(
            "webhooks",
            Webhook,
            params={k: v for k, v in params.items() if v is not None},
        )

    async def update_webhook(
        self,
        webhook_id: int,
        url: str,
        name: str | None = None,
        ids: list[str] | None = None,
        eans: list[str] | None = None,
    ):
        """Create and store a new webhook associated with the authenticated user."""
        data = {
            "url": url,
            "name": name,
            "ids": ids,
            "eans": eans,
        }
        return await self.execute(
            f"webhooks/{webhook_id}",
            Webhook,
            "patch",
            data=data,
        )

    async def delete_webhook(self, webhook_id: int):
        """Remove an existing webhook from the system."""
        return await self.execute(f"webhooks/{webhook_id}", MessageResponse, "delete")
